/* - test-PG for learning/testing -
 * urspruengliches Programm enthaelt arrays die dynamisch allociert
 * und ueber pointer referenziert werden
 * das neue Programm ersetzt (mcplx *) durch (mcplxp)
 * als c-version wie gehabt aber unter c++ als array class (mcplxp)
 * mit Uebergabe als Pointer und Arrayzugriff auf cplx[]
 * mit virtual speicher-management, blocking etc.
 * zentrale Klasse an die alle konstruktoren gemeldet werden und ueber die
 *  msg gesendet werden kann oder einfacher: instanz enthaelt von Anfang an
 *  array-pointer oder Datei-pointer
 *
 * c     v[4]    mcplx *v => a_cplx v 
 * c++   v[4]          class a_cplx v
 */
#ifndef _ARRAY_M_H
#define _ARRAY_M_H
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#define CONFIG_A_CPLX		/* define/undef */

typedef int mcplx;

#if defined(__cplusplus) && defined(CONFIG_A_CPLX)
// #define DEBUG 1
// #define NDEBUG 1
// ------------------ class.h ----------------------------------------
class a_cplx {
public:
  // Konstruktoren
  a_cplx();
  // Destruktor
  ~a_cplx(){};		// evl. Referenzen zaehlen und
			// Speicher bei Ref=0 freigeben?
  mcplx & operator[](unsigned long);	// aa[4]=7; Elementzugriff
  a_cplx& operator=(const a_cplx&);	// aa=ab;
  a_cplx& operator=(mcplx *);		// aa=b;
  unsigned long querylen(){ return length; }	// cout << a.querylen();
  void setlen(unsigned long,unsigned);

private:
  unsigned long length;		// len of array
  mcplx *thearray;		// pointer to memory (or NULL)
  // emulation of array if thearray==0
  mcplx buf;			// enthaelt ein Element
  unsigned long idx;		//  .. mit diesem index
  mcplx bufold;			// buf geaendert?
  unsigned filenum;
  FILE *f1;			// file shared, therefor pointer! :) 
};

// ------------------ class.c ----------------------------------------
inline a_cplx::a_cplx(){
  length=0;filenum=0;f1=NULL;bufold=buf=0;idx=0;
  thearray = NULL;
#if defined(DEBUG)
  printf(" %p.a_cplx",this);
#endif
}

inline void  a_cplx::setlen(unsigned long l,unsigned fnum){
#if defined(DEBUG)
   printf(" %p.setlen(%lu,%u)",this,l,fnum);
#endif
   length=l;filenum=fnum;
}

inline mcplx & a_cplx::operator[](unsigned long i){
#if defined(DEBUG)
   printf("\n %p.%p.a%d[%lu]",this,thearray,filenum,i);
#endif
#if !defined(NDEBUG)
   if(i>=length) printf("ArrayOVL i=%lu l=%lu\n",i,length);
   assert(i<length);	// !assert && !defined(NDEBUG) => ERROR
#endif
   if(thearray){
     return thearray[i];  }		/* reference! */
   else {
     if(i!=idx){
       /* write old buf if changed */
       if(bufold!=buf){
#if defined(DEBUG)
         printf(" write a%d[%lu]",filenum,idx);
#endif
         assert(!fseek(f1,idx*sizeof(buf),SEEK_SET));
         assert(fwrite(&buf,sizeof(buf),1,f1));
       }
       {
#if defined(DEBUG)
         printf(" read a%d[%lu]",filenum,i);
#endif
         assert(!fseek(f1,i*sizeof(buf),SEEK_SET));
         buf=0;fread(&buf,sizeof(buf),1,f1);
       }
       bufold=buf;
       idx=i;
     }
     return buf;			/* reference! */
   }
}

inline a_cplx& a_cplx::operator=(const a_cplx& x){	// aa=ab;
   length=x.length;
   filenum=x.filenum;
   thearray=x.thearray;
   f1=x.f1;
   buf=x.buf;
   bufold=x.bufold;
   idx=x.idx;
   return *this;
}

inline a_cplx& a_cplx::operator=(mcplx *x){	// aa=b;
   char s1[20];
#if defined(DEBUG)
   printf(" %p.thearray=%p",this,thearray);
#endif
   thearray=x;
   if(!x)if(!f1){			// virtual memory as file
     sprintf(s1,"a_cplx%d.tmp",filenum);
     f1=fopen(s1,"r+");if(!f1)f1=fopen(s1,"w+");
     assert(f1);
   }
   return *this;
}
// --------------------- end-of-class --------------------------------

// use the following definition for declarations of arrays
#define A_cplx  class a_cplx
#define A_def(x)  class a_cplx x
#define A_set_len(x,len,fnum)  x.setlen(len,fnum)
#else /* --- no C++, only c: A_cplx a,b; --- */
#define A_cplx  mcplx *
#define A_def(x)  mcplx *x
#define A_set_len(x,len,fnum)			/* do nothing */
#endif 


#if 1		/* try this as a_test.c */
int func1(A_cplx a){		/* pointer to object */
  printf("func1-test: a[3]=%d\n",a[3]);	/* korrekt ??? */
  a[3]*=2; return 0; }

int main(){
  mcplx a[5],b[8]; int i;
  A_def(pa);A_def(pb);
  pa=NULL;pb=b;A_set_len(pa,5,1);A_set_len(pb,8,2);
  printf(" sizeof(A_cplx)=%d\n",sizeof(A_cplx));
  for(i=0;i<5;i++)pa[i]=i*3;
  for(i=0;i<8;i++)pb[i]=i*2;
  func1(pa);func1(pa);
  printf( "test %d\n", pa[3]);
  printf( "test %d\n", pb[3]);
  printf( " pa=%p pb=%p &pa=%p &pb=%p\n",&(pa[0]),&(pb[0]),&pa,&pb);
  pb[8]=0; }
#endif
#endif
